/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       select.c
**     SYSTEM   NAME:       block
**     ORIGINAL AUTHOR(S):  Jan van Oorschot
**     VERSION  NUMBER:     1.00
**     CREATION DATE:       3 march 1993
**
** DESCRIPTION: This module contains a blocking library based on
**              the 'select' call of the socket library/kernel.
**              It keeps a lists of sockets which will be used
**              to block on.
**              
*************************************************************************
** CHANGES INFORMATION **
*************************
** REVISION: $Revision$
** AUTHOR:   $Author$
** DATE:     $Date$
** LOG:      $Log$
*************************************************************************/
#ifndef BSD_SELECT
#define BSD_SELECT
#endif

#ifdef OS2
#include <types.h>
#include <netlib.h>
#include <sys/select.h>       /* contains FD_*** macros */
#else
#include <sys/types.h>		
#define sock_init()	0
#define soclose(x)	close(x)
#endif
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <dnpap.h>
#include <config.h>
#include <timer.h>
#include "block.h"

struct _SocketList
{
    BLOCK *blk;
};

typedef struct _SocketList SOCKLIST;
static	SOCKLIST  	*BlockSockList;
static	int		BlockNrBlocks;
static 	int		BlockSockHigh;
static	fd_set		FDReads;
static	fd_set		FDWrites;
static	fd_set		FDExcepts;

/************************************************************************
** NAME:        BlockInit
** SYNOPSIS:    int BlockInit(void)
** DESCRIPTION: Initialises the blocking environment. It
**		should alocate the nececary resources .
**		
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int	BlockInit(void)
{
    SHORT	shrt;
    int		cnt;
    /* find out how many blocks there should be */
    if(!ConfigGetShort("block.nrblocks",&shrt))
    {
	BlockNrBlocks = 10;
    }
    else
    {
	BlockNrBlocks = (int) &shrt;
    }
    BlockSockHigh = 0;
    /* allocate the block structure */
    if ( (BlockSockList = DnpapMalloc( sizeof(SOCKLIST) * BlockNrBlocks)) == NULL)
    {
	DnpapMessage(DMC_WARNING,BLOCK_ERROR+1,"can't allocate blocks space");
	return (0 - (BLOCK_ERROR+1));
    }
    /* initiate the block structure to a well known state */
    for(cnt=0;cnt<BlockNrBlocks;cnt++)
    {
	BlockSockList[cnt].blk = NULL;
    }
    /* clear the static fd_set's */
    FD_ZERO(&FDReads);
    FD_ZERO(&FDWrites);
    FD_ZERO(&FDExcepts);
    return 0;
}
/************************************************************************
** NAME:        BlockClose
** SYNOPSIS:    int BlockClose(void)
**              
** DESCRIPTION: Closes down the blocking environment, deregistering
**		all the current blocks and freeing any memory.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int	BlockClose(void)
{
    DnpapFree(BlockSockList);
    return 0;
}

/************************************************************************
** NAME:        BlockRegister
** SYNOPSIS:    int BlockRegister(BLOCK *blk)
**		BLOCK	*blk		block structure
**              
** DESCRIPTION: Registers a new block at the blocking environment.
**		The callback function <cb> will be called every
**		time the block is triggered. <cb> is defined
**		in the BLOCK structure.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int BlockRegister(BLOCK *blk)
{
    /*check if <blk> contains a valid  socket number */
    if(blk->sckt < 0 || blk->sckt>= BlockNrBlocks)
    {
	DnpapMessage(DMC_WARNING,BLOCK_ERROR+2,"wrong socketnumber");
	return(0 - (BLOCK_ERROR+2));
    }
    /* check if this socket is already registered */
    if(BlockSockList[blk->sckt].blk != NULL)
    {
	DnpapMessage(DMC_WARNING,BLOCK_ERROR+3,"block already registered");
	return(0 - (BLOCK_ERROR+3));
    }
    /* check if this socket has the highest port number */
    if(BlockSockHigh <= blk->sckt)
    {
	BlockSockHigh = blk->sckt + 1;
    }
    /* add it to the list */
    BlockSockList[blk->sckt].blk = blk;
    /* mark this new socket in the static fs_set's */
    FD_SET(blk->sckt,&FDReads);
    FD_SET(blk->sckt,&FDWrites);
    FD_SET(blk->sckt,&FDExcepts);
    
    return 0;
}
/************************************************************************
** NAME:        BlockDeRegister
** SYNOPSIS:    int BlockDeRegister(BLOCK *blk)
**		BLOCK *blk;		block structure
**              
** DESCRIPTION: Removes the block <blk> from the blocking
**		environment. This should always be possible
**		since we evidently are currently not blocked.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int BlockDeRegister(BLOCK *blk)
{
    int tmpmax,cnt;
    /* check if this was the highest socket, and if 
       so, find the new highest */
    if(blk->sckt == (BlockSockHigh-1))
    {
	tmpmax = 0;
	for(cnt=0;cnt<BlockNrBlocks;cnt++)
	{
	    if(BlockSockList[cnt].blk !=NULL)
	    {
		if(BlockSockList[cnt].blk->sckt > tmpmax)
		{
		    tmpmax = BlockSockList[cnt].blk->sckt;
		}
		else
		{
		    	DnpapMessage(DMC_WARNING,BLOCK_ERROR+3,"internal error");
			return(0 - (BLOCK_ERROR+3));
		}
	    }
	}
	BlockSockHigh = tmpmax+1;
    }
    /* remove it from list */
    BlockSockList[blk->sckt].blk = NULL;

    /* remove the the bit from the fd_set's */ 
    FD_CLR(blk->sckt,&FDReads);
    FD_CLR(blk->sckt,&FDWrites);
    FD_CLR(blk->sckt,&FDExcepts);
    return 0;
}
/************************************************************************
** NAME:        Block
** SYNOPSIS:    int Block(void);
**              
** DESCRIPTION: Blocks until one of the block's is released.
**		The callback function of this block will be called
**		with a reason code. 
**		In a normal environment, this function will be put
**		in a tight loop.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int Block(void)
{
    fd_set fd_read,fd_write,fd_except;
    LONG t;
    struct timeval timeout;
    int iret,cnt;

    /* copy static fd_set's to the local copies */
    FD_ZERO(&fd_read);
    FD_ZERO(&fd_write);
    FD_ZERO(&fd_except);
    for(cnt=0;cnt<BlockSockHigh;cnt++)
    {
	if(BlockSockList[cnt].blk!=NULL)
	{
		FD_SET(cnt,&fd_read);
		FD_SET(cnt,&fd_write);
		FD_SET(cnt,&fd_except);
	}
    }
    /* request the next timeout from the timer library and do select */
    if((t = TimerRemaining()) == -1L )
    {
    	iret = select(BlockSockHigh,&fd_read,NULL,&fd_except,NULL);
    }
    else
    {
    timeout.tv_sec = t/1000;
    timeout.tv_usec = (t%1000)*1000;
    iret = select(BlockSockHigh,&fd_read,NULL,&fd_except,&timeout);
    }

    /* select reason: error */
    if(iret < 0)
    {
    	perror("");
    	DnpapMessage(DMC_WARNING,BLOCK_ERROR+4,"select error");
	return(0 - (BLOCK_ERROR+4));
    }
    /* select reason: timeout */
    if(iret == 0)
    {
	if(!TimerCheck())
	{
	    DnpapMessage(DMC_WARNING,BLOCK_ERROR+5,"timout without reason");
	}
	return 0;
    }
    /* select reason: socket event */
    /* for all sockets, check if something has happend */
    for(cnt=0;cnt<BlockSockHigh;cnt++)
    {
	if(BlockSockList[cnt].blk!=NULL)
	{
/*
	    if(FD_ISSET(BlockSockList[cnt].blk->sckt,&fd_write))
	    {
		(BlockSockList[cnt].blk->cb)(BlockSockList[cnt].blk,BLOCKEVENT_WRITE);
	    }
*/
	    if(FD_ISSET(BlockSockList[cnt].blk->sckt,&fd_except))
	    {
		(BlockSockList[cnt].blk->cb)(BlockSockList[cnt].blk,BLOCKEVENT_EXCEPT);
	    }
	    if(FD_ISSET(BlockSockList[cnt].blk->sckt,&fd_read))
	    {
		(BlockSockList[cnt].blk->cb)(BlockSockList[cnt].blk,BLOCKEVENT_READ);
	    }
	}
    }
    return 0;
}
/************************************************************************
** NAME:        BlockCheck
** SYNOPSIS:    int BlockCheck(void)
**              
** DESCRIPTION: Does the same as Block, but without the actual 
**		blocking. Should seldom be used, since it
**		implements busy-waiting like behaviour.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int BlockCheck(void)
{
    return 0;
}
