/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       nwapi.c
**     SYSTEM   NAME:       Game
**     ORIGINAL AUTHOR(S):  Jan van Oorschot
**     VERSION  NUMBER:     1.00
**     CREATION DATE:       5 march 1993
**
** DESCRIPTION: The access functions for the network layer. This layer
**		forms an abstraction of the services offered by
**		the network protocols used like tcp/ip and serial line
**		networking.
**
**		Services offered are
**		. abstraction of the support network protocols
**		. initialisation
**		. the sending of a frame over a NWCONN. Such
**		  a send is blocking.
**		. creation request for a new connection
**		. callback function for the receiving of frames
**		  over a connection. (connection specific)
**		. callback function for the notification
**		  of the creation of a new NWCONN (one for
**		  the entire network layer)
**		. callback function for the notificationCON
**		  of the termination of a connection (one for
**		  the entire network layer)
**              
** STRUCTURES:	
**		NWCONN		; structure per interface
**			. family
**			. network connection union
**			. connection info used during Open
**			. reception callback
**			. shutdown callback
**			. statistics
**
**		IFACES			; structure per interface
**			. family-id
**			. send-hook
**			. init-hook
**			. closedown hook
**			. 
**
** FUNCTIONS:
**		NWInit		; opens and establishes connection
**				; create/close callbacks
**		NWOpen		; creates a NWCONN, establishes
**				; a receive-callback
**		NWClose		; closes NWCONN
**		NWSend		; sends a message through the NWCONN
**				; and blocks until finished
*************************************************************************
** CHANGE INFORMATION **
*************************
** REVISION:    $Revision$
** WORKFILE:    $Workfile$
** LOGINFO:     $Log$
*************************************************************************/
#include	<stdio.h>
#include	<string.h>
#include	<dnpap.h>

#include	"nw.h"

#define IFACE_TCP 
#define IFACE_UDP
#define IFACE_NIT

#include <block.h>

#ifdef IFACE_TCP
#include "nwtcp/nwtcp.h"
#endif
#ifdef IFACE_UDP
#include "nwudp/nwudp.h"
#endif
#ifdef IFACE_NIT
#include "nwnit/nwnit.h"
#endif

typedef struct _nwiface NWIFACE;

struct _nwiface
{
    char *IfaceName;
    int (*IfaceInit)(
		       int (*CBOpen)(NWCONN *,char *),
		       int (*CBClose)(NWCONN *),
		       int (*CBDgramRec)(NWCONN *,BYTE *,int),
	               void *param
		       );
    int (*IfaceExit)(void);
    NWCONN * (*IfaceOpen)(
		       char *ConnInfo,
		       int (*CBReceive)(NWCONN *con,BYTE *frame,int len)
		       );  
    int (*IfaceClose)(NWCONN *conn);
    int (*IfaceSend)(
		       NWCONN *conn,
		       BYTE *frame,
		       int len);
    void (*IfaceSetRecCB)(
		       NWCONN *conn,
		       int (*CBReceive)(NWCONN *con,BYTE *frame,int len)
		       );  
    char *(*IfaceEntry)(void);
    int (*IfaceDgramSend)(char *ConnInfo,BYTE *frame,int len);
    int (*IfaceDgramRepl)(NWCONN *con,BYTE *frame,int len);
    int (*IfaceConfig)(char *cfg);
};


/* if you add one, remember to adjust the Iface defines in nw.h */
NWIFACE	ifaces[] =
{
#ifdef IFACE_TCP
    {	"tcp",		TCPInit,	TCPExit,	TCPOpen,
	 TCPClose,	TCPSend,	TCPSetRecCB,	TCPGetEntry,
         NULL, 		NULL,		TCPConfig
    },
#endif
#ifdef IFACE_UDP
    {	"udp",		UDPInit,	UDPExit,	UDPOpen,   
	UDPClose,   	UDPSend,	NULL,		UDPGetEntry,
        UDPDgramSend,	UDPDgramRepl,	UDPConfig
    },
#endif
#ifdef IFACE_NIT
    {	"nit",		NwNitInit,	NwNitExit,	NwNitOpen,
	NwNitClose,	NwNitSend,	NwNitSetRecCB,	NwNitGetEntry,
        NULL,		NULL,		NULL
    },
#endif
    {	NULL, 		NULL,   	NULL,   	NULL,   
	NULL,    	NULL,   	NULL,       	NULL,
        NULL,		NULL,		NULL
    }
 };

/* static variables */
static int (*NWCBOpen)(NWCONN *con,char *istr);
static int (*NWCBClose)(NWCONN *con);
static int (*NWCBDgramRec)(NWCONN *con,BYTE *frame,int len);

/* prototypes of NW callback functions */
static int CBNWOpen(NWCONN *con,char *info);
static int CBNWClose(NWCONN *con);
static int CBNWReceive(NWCONN *con,BYTE *frame,int len);
static int CBNWDgramRec(NWCONN *con,BYTE *frame,int len);

/************************************************************************
** NAME:        NWInit
** SYNOPSIS:    int NWInit(CBAppConnOpen,CBAppConnClose,CBAppDgramRec)
**              int (*CBAppConnOpen)(NWCONN *con, char*istr);
**		int (*CBAppConnClose)(NWCONN *con);
**		int (*CBAppDgramRec)(NWCONN *con, BYTE *frame,int len);
**              
** DESCRIPTION: Initiates the NW layer. The application
**		passes two callback functions to the network
**		layer which will be called when a connection
**		in created or closed by a calling node.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int NWInit( int (*CBAppConnOpen)(NWCONN *con,char *istr),
	    int (*CBAppConnClose)(NWCONN *con),
	    int (*CBAppDgramRec)(NWCONN *con,BYTE *frame,int len))
{
    int cnt,iret;
    /* remember application callback functions */
    NWCBOpen = CBAppConnOpen;
    NWCBClose = CBAppConnClose;
    NWCBDgramRec = CBAppDgramRec;
    /* initiate the lower layers */
    cnt = 0;
    while(ifaces[cnt].IfaceInit != NULL)
    {
	if (( iret = ifaces[cnt].IfaceInit(
	    CBNWOpen,
	    CBNWClose,
	    CBNWDgramRec,
	    NULL)) != 0)
	{
	    DnpapMessage(DMC_WARNING,NW_ERROR+7,"init failed");
	    return NW_ERROR+7;
	}
	cnt++;
    }
    return 0;
}
/************************************************************************
** NAME:        NWExit
** SYNOPSIS:    int NWExit()
**              
** DESCRIPTION: Closes down the network layer, closing all
**		open connections
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int	NWExit(void)
{
    int cnt,iret;
    /* initiate the lower layers */
    cnt = 0;
    while(ifaces[cnt].IfaceExit != NULL)
    {
	if (( iret = ifaces[cnt].IfaceExit()) != 0)
	{
	    DnpapMessage(DMC_WARNING,NW_ERROR+7,"exit failed");
	    return NW_ERROR+7;
	}
	cnt++;
    }
    return 0;
}
/************************************************************************
** NAME:        NWOpen
** SYNOPSIS:    NWCONN *NWOpen(ConnInfo,CBAppReceive);
**              char	ConnInfo;	complete connection info
**					example: "tcp:dnpap.et:1024"
**              int	(*CBAppReceive)(NWCONN *con, BYTE *frame,int len);
** DESCRIPTION: Opens a connection. <ConnInfo> should contain 
**		all information needed to open a connection 
**		and has the following format:
**
**			<nw-type>:<connection-info>
**
**		If a frame is received over this connection,
**		the callback function <CBAppReceive> will
**		be called to inform the appliction of the
**		receiped.
** RETURNS:     NULL--> open failed
**              else    Network connection structure
** RETURNS:     void
*************************************************************************/
NWCONN	*NWOpen(
		char *ConnInfo,
		int (*CBAppReceive)(NWCONN *con,BYTE *frame,int len))
{
    char id[32],*cptr;
    int cnt;
    NWCONN *con;
    /* collect the iface identifier and search for iface */
    cptr = ConnInfo;
    cnt =0;
    while(*cptr != ':' && *cptr != '\0')
    {
	id[cnt++]=*cptr++;
    }
    if(*cptr =='\0')
    {
	return NULL;
    }
    id[cnt] = '\0';
    cptr++;
    cnt = 0;
    while(ifaces[cnt].IfaceName != NULL)
    {
	if(strcmp(ifaces[cnt].IfaceName,id) == 0)
	    break;
	cnt ++;
    }
    if(ifaces[cnt].IfaceName == NULL)
	return NULL;
    /* call correct open function */
    if(ifaces[cnt].IfaceOpen != NULL)
    {
	if((con = (ifaces[cnt].IfaceOpen)(cptr,CBNWReceive)) == NULL)
	{
	    return NULL;
	}
    }
    else
	return NULL;
    con->nwinfo.type = cnt;
    con->nwinfo.CBAppReceive = CBAppReceive;
    con->nwinfo.ConnInfo = DnpapMalloc(strlen(ConnInfo));
    strcpy(con->nwinfo.ConnInfo,ConnInfo);
    return con;
}
	    
/************************************************************************
** NAME:        NWClose
** SYNOPSIS:    int NWClose(NWCONN *conn)
**              
** DESCRIPTION: Closes a connection <conn>
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int	NWClose(NWCONN *conn)
{	
    if(ifaces[conn->nwinfo.type].IfaceClose != NULL)
	return (ifaces[conn->nwinfo.type].IfaceClose)(conn);
    else
	return 1;
}
/************************************************************************
** NAME:        NWSend
** SYNOPSIS:    int NWSend(NWCONN *conn,BYTE *frame, int len)
**              
** DESCRIPTION: Sends frame <frame> of length <len> over 
**		connection <conn>.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int	NWSend(NWCONN *conn,BYTE *frame, int len)
{
    if(ifaces[conn->nwinfo.type].IfaceSend != NULL)
	return (ifaces[conn->nwinfo.type].IfaceSend)(conn,frame,len);
    else
	return 1;
}
/************************************************************************
** NAME:        NWSetReceiveCB
** SYNOPSIS:    void NWSetReceiveCB(
**			NWCONN *conn,
**			int (*CBAppReceive)(NWCONN *,BYTE *,int))
**
** DESCRIPTION: sets the application receive callback function
**		of a connection. This function will be used when
**		a connection is initiated by an other node.
** RETURNS:     void
*************************************************************************/
void NWSetReceiveCB(
	    NWCONN *conn, 
	    int (*CBAppReceive)(NWCONN *conn, BYTE *frame,int len))
{
    conn->nwinfo.CBAppReceive = CBAppReceive;
}
/************************************************************************
** NAME:        NWGetEntries
** SYNOPSIS:    int NWGetEntries(char * buf, int len);
** DESCRIPTION: Stores the entries of all interfaces in <buf>
**		and takes care that this buffer does not overflow.
**		the format of <buf> will be:
**
**			<entry>\n<entry>\n.....<entry>\0
** RETURNS:     number of entries in <buf>
*************************************************************************/
int NWEntries(char *buf,int len)
{
    int cnt = 0,lcnt,tlen;
    char *cptr;

    lcnt = len;
    buf[0]= '\0';
    while(ifaces[cnt].IfaceName != NULL)
    {
	cptr = (ifaces[cnt].IfaceEntry)();
	if((tlen=strlen(cptr)) > lcnt)
	    return cnt;
	strcat(buf,cptr);
	strcat(buf,"\n");
	lcnt -= tlen+1;
	cnt ++;
    }
    return cnt;
}
/************************************************************************
** NAME:        NWDgramSend
** SYNOPSIS:    int NWDgramSend(char *to,BYTE *frame, int len);
** DESCRIPTION: Sends a datagram to the destination given by <to>
**		The format of <to> should be something like:
**
**			udp:dutepp0:1024
**
** RETURNS:     0 ---> correct
**		  else errorcode
*************************************************************************/
int NWDgramSend(char *ConnInfo, BYTE *frame, int len)
{
    char id[32],*cptr;
    int cnt;
    /* collect the iface identifier and search for iface */
    cptr = ConnInfo;
    cnt =0;
    while(*cptr != ':' && *cptr != '\0')
    {
	id[cnt++]=*cptr++;
    }
    if(*cptr =='\0')
    {
	return 1;
    }
    id[cnt] = '\0';
    cptr++;
    cnt = 0;
    while(ifaces[cnt].IfaceName != NULL)
    {
	if(strcmp(ifaces[cnt].IfaceName,id) == 0)
	    break;
	cnt ++;
    }
    if(ifaces[cnt].IfaceName == NULL)
	return 2;
    /* call correct dgram funcion */
    if(ifaces[cnt].IfaceDgramSend != NULL)
	return (ifaces[cnt].IfaceDgramSend)(cptr,frame,len);
    else
	return 3;
}
/************************************************************************
** NAME:        NWDgramRepl
** SYNOPSIS:    int NWDgramRepl(NWCONN *con,BYTE *frame, int len);
** DESCRIPTION: Sends a reply datagram to the destination given by <con>
**		The call can only be made from within 
**		the DgramRec callback function. <con> should be
**		a NWCONN structure that is passed to the
**		application by that callback function.
**		For normal datagram sends, use NWDgramSend.
** RETURNS:     0 ---> correct
**		  else errorcode
*************************************************************************/
int NWDgramRepl(NWCONN *con, BYTE *frame, int len)
{
    if(ifaces[con->nwinfo.type].IfaceDgramRepl != NULL)
	return (ifaces[con->nwinfo.type].IfaceDgramRepl)(con,frame,len);
    else
	return 1;
}
/************************************************************************
** NAME:        NWConfig
** SYNOPSIS:    NWCONN *NWConfig(ConfInfo);
**              char	ConfInfo;	complete config info
**					example: "tcp:noserver"
** DESCRIPTION: Configs a connection. <ConnInfo> should contain 
**		points to a config string.
**		and has the following format:
**
**			<nw-type>:<config-info>
**
** RETURNS:     NULL--> open failed
**              else    Network connection structure
** RETURNS:     void
*************************************************************************/
int NWConfig(char *ConfInfo)
{
    char id[32],*cptr;
    int cnt;

    /* collect the iface identifier and search for iface */
    cptr = ConfInfo;
    cnt =0;
    while(*cptr != ':' && *cptr != '\0')
    {
	id[cnt++]=*cptr++;
    }
    if(*cptr =='\0')
    {
	return NULL;
    }
    id[cnt] = '\0';
    cptr++;
    cnt = 0;
    while(ifaces[cnt].IfaceName != NULL)
    {
	if(strcmp(ifaces[cnt].IfaceName,id) == 0)
	    break;
	cnt ++;
    }
    if(ifaces[cnt].IfaceName == NULL)
	return 1;
    /* call correct open function */
    if(ifaces[cnt].IfaceConfig != NULL)
    {
	return (ifaces[cnt].IfaceConfig)(cptr);
    }
    else
	return 0;
}
/* NW callback functions */
static int CBNWOpen(NWCONN *con,char *ConnInfo)
{
    char *cptr,id[32];
    int cnt;
    /* called when a connection is opened by the other side,
       the <con> structure is not completly initialised.
       Do this now, and inform the application layer of
       the creation */
    /* collect the iface identifier and search for iface */
    cptr = ConnInfo;
    cnt =0;
    while(*cptr != ':' && *cptr != '\0')
    {
	id[cnt++]=*cptr++;
    }
    if(*cptr =='\0')
    {
	return 1;
    }
    id[cnt] = '\0';
    cnt = 0;
    while(ifaces[cnt].IfaceName != NULL)
    {
	if(strcmp(ifaces[cnt].IfaceName,id) == 0)
	    break;
	cnt ++;
    }
    if(ifaces[cnt].IfaceName == NULL)
	return 1;
    con->nwinfo.type = cnt;

    /* collect the iface identifier and search for iface */
    cptr = ConnInfo;
    cnt =0;
    while(*cptr != ':' && *cptr != '\0')
    {
	id[cnt++]=*cptr++;
    }
    if(*cptr =='\0')
    {
	return 1;
    }
    id[cnt] = '\0';
    cnt = 0;
    while(ifaces[cnt].IfaceName != NULL)
    {
	if(strcmp(ifaces[cnt].IfaceName,id) == 0)
	    break;
	cnt ++;
    }
    if(ifaces[cnt].IfaceName == NULL)
	return 1;
    /*let interface set the correct receive callback to this layer */
    (ifaces[cnt].IfaceSetRecCB)(con,CBNWReceive);

    /* inform the application of the open */
    if(NWCBOpen != NULL)
	return NWCBOpen(con,ConnInfo);
    else
	return 1;
}
static int CBNWClose(NWCONN *con)
{
    /* inform application of the close */
    if(NWCBClose != NULL)
	return NWCBClose(con);
    else
	return 1;
}
static int CBNWReceive(NWCONN *con,BYTE *frame,int len)
{
    /* inform application of the receiped */
    return (con->nwinfo.CBAppReceive)(con,frame,len);
}

static int CBNWDgramRec(NWCONN *con,BYTE *frame,int len)
{
    /* call the Datagram Receive callback of the application,
       if this is available */
    if(NWCBDgramRec != NULL)
    {
	return NWCBDgramRec(con,frame,len);
    }
    else
	return 1;
}
